function [outputs] = SMC_PB_for_GC_content(y, inputs, z)

% [outputs] = SMC_PB_for_GC_content(y, inputs, z)
%
% Simplified version of SMC_PB_for_GC_content. Bootstrap for filtering and
% multinomial for resampling only.
%

% Extract the initial values:
alpha = inputs.init_params.alpha;
beta = inputs.init_params.beta;
eta = inputs.init_params.eta;
delta = inputs.init_params.delta;
success_prob = inputs.init_params.success_prob;
mdl_trans_mtx = inputs.init_params.mdl_trans_mtx;

Q = length(eta); T = length(y);
% The vector of indicator for missing data problem
if nargin == 2
    z = ones(1,T);
end

model_params = cell(1,Q);
for q = 1:Q
    
    seg_length_pmf_name = 'geometric_pdf';
    seg_length_pmf_params.success_prob = success_prob(q);
    seg_length_pmf_params.start_pnt = 1;
    
    data_log_lkl_pdf_name = 'log_lkl_GC_content';
    
    % Inner parameters for the GC_content system:
    inner_params.alpha = alpha;
    inner_params.beta = beta;
    inner_params.delta = delta(q);
    inner_params.eta = eta(q);
    
    model_params{q}.seg_length_pmf_name = seg_length_pmf_name;
    model_params{q}.seg_length_pmf_params = seg_length_pmf_params;
    model_params{q}.data_log_lkl_pdf_name = data_log_lkl_pdf_name;
    model_params{q}.inner_params = inner_params;
    
end

% Extract the system parameters:
N = inputs.Num_of_pars;
system_type = inputs.system_type;

% Initialize particles:
current_particles(N) = struct;
for i = 1:N
    current_particles(i).D = 0; current_particles(i).M = mod(i-1, Q) + 1;
end

% Initialize the log-filtering density
log_filt_weights = -log(N)*ones(1, N);
filt_weights = exp(log_filt_weights);

% Calculate the parameters of the posterior distribution of the latent
% variable at time t=1.
alpha_post_curr = zeros(Q, 1);
beta_post_curr = zeros(Q, 1);
eta_post_curr = zeros(Q, 1);
delta_post_curr = zeros(Q, 1);
for q = 1:Q
    prior_params.alpha = alpha;
    prior_params.beta = beta;
    prior_params.eta = eta(q);
    prior_params.delta = delta(q);
    
    % Posterior dist. conditioned on the current particle at time t
    [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
        y(1), z(1));
    alpha_post_curr(q) = post_params_curr.alpha_post;
    beta_post_curr(q) = post_params_curr.beta_post;
    eta_post_curr(q) = post_params_curr.eta_post;
    delta_post_curr(q) = post_params_curr.delta_post;
end

%% Initalize the intermediate quantities needed for on-line EM
% The quantities for parameter of the geometric distribution:
S_1 = zeros(N, Q); S_2 = zeros(N, Q);
% The quantities for the transition probability matrix for the models:
S_3 = ones(N, Q, Q);
% The quantities for alpha and beta:
S_4 = psi(alpha_post_curr(mod((0:(N-1))', Q)+1)) ...
    - log(beta_post_curr(mod((0:(N-1))', Q)+1));
% The quantites for delta:
S_5 = zeros(N, Q);
S_6 = zeros(N, Q); S_7 = zeros(N, Q);
for i = 1:N
    q = current_particles.M;
    S_1(i, q) = 1;
    S_5(i, q) = 1/delta_post_curr(q) ...
        + eta_post_curr(q)^2*alpha_post_curr(q)/beta_post_curr(q);
    S_6(i, q) = eta_post_curr(q)*(alpha_post_curr(q)/beta_post_curr(q));
    S_7(i, q) = (alpha_post_curr(q)/beta_post_curr(q));
end

%% Initialise the sufficient statistics and store them

SS = zeros(T, Q^2 + 5*Q + 1);
for q1 = 1:Q
    for q2 = 1:Q
        SS(1, 2*Q + (q1-1)*Q + q2) = filt_weights*S_3(:,q1,q2);
    end
end
SS(1, [1:(2*Q) (Q^2 + 2*Q + 1):(Q^2 + 5*Q + 1)]) = ...
    filt_weights*[S_1 S_2 S_4 S_5 S_6 S_7];

%% Recursion over time:
for t = 2:T
    
    if mod(t, 100) == 0
        fprintf('%d ', t);
        if mod(t, 1000) == 0
            fprintf('\n');
        end
    end 
    prev_particles = current_particles;   
    
    %%%%%%%%%%%%%%%%%%%%% Bootstrap Filtering %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Bootstrap method:
    cand_particles = prev_particles;
    log_cand_weights = zeros(1, N);
    
    for i = 1:N
        % Sample from the Proposal density
        M_prev = prev_particles(i).M;
        D_prev = prev_particles(i).D;
        if rand < success_prob(M_prev)
            cand_particles(i).D = t-1;
            cand_particles(i).M = randgen(1:Q, 1, mdl_trans_mtx(M_prev, :));
            % Incremental weight:
            inc_weight = log_lkl_y(y(t), z(t), model_params{cand_particles(i).M}, system_type);
        else
            % Incremental weight:
            inc_weight = log_lkl_y(y(D_prev +1:t), z(D_prev +1:t),model_params{M_prev}, system_type)...
                - log_lkl_y(y(D_prev+1:t-1), z(D_prev+1:t-1), model_params{M_prev}, system_type);
        end
        log_cand_weights(i) = inc_weight;
    end
    log_cand_weights = log_cand_weights - log_sum(log_cand_weights);
    cand_weights = exp(log_cand_weights);

    
    %%%%%%%%%%%%%%%%%%%%%%%% Sufficient Statistics update %%%%%%%%%%%%%%%%%
    for i = 1:N      
        cd = cand_particles(i).D;
        cm = cand_particles(i).M;      
        pd = prev_particles(i).D;
        pm = prev_particles(i).M;
        
        prior_params.alpha = alpha;
        prior_params.beta = beta;
        prior_params.eta = eta(cm);
        prior_params.delta = delta(cm);
        
        % Posterior dist. conditioned on the current particle at time t
        [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t), z(cd+1:t));
        alpha_post_curr = post_params_curr.alpha_post;
        beta_post_curr = post_params_curr.beta_post;
        eta_post_curr = post_params_curr.eta_post;
        delta_post_curr = post_params_curr.delta_post;
        
        % Posterior dist. conditioned on the current particle at time t-1
        [post_params_prev] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t-1), z(cd+1:t-1));
        alpha_post_prev = post_params_prev.alpha_post;
        beta_post_prev = post_params_prev.beta_post;
        eta_post_prev = post_params_prev.eta_post;
        delta_post_prev = post_params_prev.delta_post;
        
        if cd == t-1                        
            S_1(i, cm) = S_1(i, cm) + 1; 
            S_3(i, pm, cm) = S_3(i, pm, cm) + 1;
            % Alpha and beta
            S_4(i) = S_4(i) + psi(alpha_post_curr) - log(beta_post_curr);
            % delta and eta
            S_5(i, cm) = S_5(i, cm) + 1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr;
            S_6(i, cm) = S_6(i, cm) + eta_post_curr*(alpha_post_curr/beta_post_curr);
            S_7(i, cm) = S_7(i, cm) + (alpha_post_curr/beta_post_curr);
        elseif (cd == pd)&&(cm == pm)
            S_2(i, cm) = S_2(i, cm) + 1;
            % Alpha and beta
            S_4(i) = S_4(i) + psi(alpha_post_curr) - log(beta_post_curr)...
                - (psi(alpha_post_prev) - log(beta_post_prev));
            % delta and eta
            S_5(i, cm) = S_5(i, cm) + 1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr...
                - (1/delta_post_prev + eta_post_prev^2*alpha_post_prev/beta_post_prev);
            S_6(i, cm) = S_6(i, cm) + eta_post_curr*(alpha_post_curr/beta_post_curr)...
                - eta_post_prev*(alpha_post_prev/beta_post_prev);
            S_7(i, cm) = S_7(i, cm) + (alpha_post_curr/beta_post_curr - alpha_post_prev/beta_post_prev);
        end
    end
    
    %% Update the sufficient statistics
    for q1 = 1:Q
        for q2 = 1:Q
            SS(t, 2*Q + (q1-1)*Q + q2) = cand_weights*S_3(:,q1,q2);
        end
    end
    SS(t, [1:(2*Q) (Q^2 + 2*Q + 1):(Q^2 + 5*Q + 1)]) = ...
        cand_weights*[S_1 S_2 S_4 S_5 S_6 S_7];
    
    %%%%%%%%%%%%%%%%% Resampling to N particles %%%%%%%%%%%%%%%%%%%%%%%%%%%
    resampled_indices = multinomial_res(log_cand_weights, N);
    current_particles = cand_particles(resampled_indices);
    S_1 = S_1(resampled_indices, :);
    S_2 = S_2(resampled_indices, :);
    S_3 = S_3(resampled_indices, :, :);
    S_4 = S_4(resampled_indices);
    S_5 = S_5(resampled_indices, :);
    S_6 = S_6(resampled_indices, :);
    S_7 = S_7(resampled_indices, :);
end

%% Outputs
outputs.SS = SS;