function [outputs] = online_EM_FS_for_GC_content(y, inputs, z)

% [outputs] = online_EM_FS_for_GC_content(y, inputs, z)
%
% This function is a simplified version of online_EM_for_GC_content. 
% Bootstrap for filtering and multinomial for resampling are used.
%


% Extract the initial values:
alpha = inputs.init_params.alpha;
beta = inputs.init_params.beta;
eta = inputs.init_params.eta;
delta = inputs.init_params.delta;
success_prob = inputs.init_params.success_prob;
mdl_trans_mtx = inputs.init_params.mdl_trans_mtx;

Q = length(eta); T = length(y);

% Initialize the arrays of parameter estimates over time:
Alpha_est = zeros(1, T); Alpha_est(1) = alpha;
Beta_est = zeros(1, T); Beta_est(1) = beta;
Delta_est = zeros(Q, T); Delta_est(:,1) = delta;
Eta_est = zeros(Q, T); Eta_est(:,1) = eta;
Success_prob_est = zeros(Q, T); Success_prob_est(:,1) = success_prob;
Mdl_trans_mtx_est = zeros(Q^2, T); Mdl_trans_mtx_est(:,1) = mdl_trans_mtx(:);

% The vector of indicator for missing data problem
if nargin == 2
    z = ones(1,T);
end

model_params = cell(1,Q);
for q = 1:Q

    seg_length_pmf_name = 'geometric_pdf';
    seg_length_pmf_params.success_prob = success_prob(q);
    seg_length_pmf_params.start_pnt = 1;

    data_log_lkl_pdf_name = 'log_lkl_GC_content';

    % Inner parameters for the GC_content system:
    inner_params.alpha = alpha;
    inner_params.beta = beta;
    inner_params.delta = delta(q);
    inner_params.eta = eta(q);

    model_params{q}.seg_length_pmf_name = seg_length_pmf_name;
    model_params{q}.seg_length_pmf_params = seg_length_pmf_params;
    model_params{q}.data_log_lkl_pdf_name = data_log_lkl_pdf_name;
    model_params{q}.inner_params = inner_params;
end

% Extract the system parameters:
N = inputs.Num_of_pars;
burn_in_per = inputs.burn_in_per;
tilda = inputs.tilda;
system_type = inputs.system_type;

% Initialize the vector for the number of particles:
N_reorg = zeros(1, T);
N_cand_reorg = zeros(1, T);

% Initialize particles:
current_particles(N) = struct;
for i = 1:N
    current_particles(i).D = 0; current_particles(i).M = mod(i-1, Q) + 1;
end

% Initialize the log-filtering density
log_filt_weights = -log(N)*ones(1, N);

% Reorganise the particles so that the particles with same values gather
[particles_reorg, log_filt_weights_reorg] ...
    = reorganize_particles(current_particles, log_filt_weights);
N_reorg(1) = length(log_filt_weights_reorg);

% Calculate the parameters of the posterior distribution of the latent
% variable at time t=1.
alpha_post_curr = zeros(Q, 1);
beta_post_curr = zeros(Q, 1);
eta_post_curr = zeros(Q, 1);
delta_post_curr = zeros(Q, 1);
for q = 1:Q
    prior_params.alpha = alpha;
    prior_params.beta = beta;
    prior_params.eta = eta(q);
    prior_params.delta = delta(q);

    % Posterior dist. conditioned on the current particle at time t
    [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
        y(1), z(1));
    alpha_post_curr(q) = post_params_curr.alpha_post;
    beta_post_curr(q) = post_params_curr.beta_post;
    eta_post_curr(q) = post_params_curr.eta_post;
    delta_post_curr(q) = post_params_curr.delta_post;
end

%% %%%%%%%%%%% Initalization for the forward smoothing method
% The quantities for parameter of the geometric distribution:
T_1 = zeros(N_reorg(1), Q); T_2 = zeros(N_reorg(1), Q);
% The quantities for the transition probability matrix for the models:
T_3 = ones(N_reorg(1), Q, Q);
% The quantities for alpha and beta:
T_4 = psi(alpha_post_curr(mod((0:(N_reorg(1)-1))', Q)+1)) ...
    - log(beta_post_curr(mod((0:(N_reorg(1)-1))', Q)+1));
% The quantites for delta:
T_5 = zeros(N_reorg(1), Q);
T_6 = zeros(N_reorg(1), Q); T_7 = zeros(N_reorg(1),Q);
for i = 1:N_reorg(1)
    q = particles_reorg.M;
    T_1(i, q) = 1;
    T_5(i, q) = 1/delta_post_curr(q) ...
        + eta_post_curr(q)^2*alpha_post_curr(q)/beta_post_curr(q);
    T_6(i, q) = eta_post_curr(q)*(alpha_post_curr(q)/beta_post_curr(q));
    T_7(i, q) = (alpha_post_curr(q)/beta_post_curr(q));
end

% Initialise the sufficient statistics and store them
filt_weights_reorg = exp(log_filt_weights_reorg);
SS_FS = zeros(T, Q^2 + 5*Q + 1);
for q1 = 1:Q
    for q2 = 1:Q
        SS_FS(1, 2*Q + (q1-1)*Q + q2) = filt_weights_reorg*T_3(:,q1,q2);
    end
end
SS_FS(1, [1:(2*Q) (Q^2 + 2*Q + 1):(Q^2 + 5*Q + 1)]) = ...
    filt_weights_reorg*[T_1 T_2 T_4 T_5 T_6 T_7];

%%%%%%%%%%%%%%%%%%%%%%%%%% Recursion over time: %%%%%%%%%%%%%%%%%%%%%%%%%%%
for t = 2:T

    if mod(t, 100) == 0
        fprintf('%d ', t);
        if mod(t, 1000) == 0
            fprintf('\n');
        end
    end
    % Step size
    if t <= burn_in_per
        gam = 0.01;
    else
        gam = (t)^(-tilda);
    end
    w1 = gam; w2 = 1 - gam;

    prev_particles = current_particles;
    log_filt_weights_reorg_prev = log_filt_weights_reorg;
    particles_reorg_prev = particles_reorg;

    %%%%%%%%%%%%%%%%%%%%% Bootstrap Filtering %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Bootstrap method:
    cand_particles = prev_particles;
    log_cand_weights = zeros(1, N);

    for i = 1:N
        % Sample from the Proposal density
        M_prev = prev_particles(i).M;
        D_prev = prev_particles(i).D;
        if rand < success_prob(M_prev)
            cand_particles(i).D = t-1;
            cand_particles(i).M = randgen(1:Q, 1, mdl_trans_mtx(M_prev, :));
            % Incremental weight:
            inc_weight = log_lkl_y(y(t), z(t), model_params{cand_particles(i).M}, system_type);
        else
            % Incremental weight:
            inc_weight = log_lkl_y(y(D_prev +1:t), z(D_prev +1:t),model_params{M_prev}, system_type)...
                - log_lkl_y(y(D_prev+1:t-1), z(D_prev+1:t-1), model_params{M_prev}, system_type);
        end
        log_cand_weights(i) = inc_weight;
    end
    log_cand_weights = log_cand_weights - log_sum(log_cand_weights);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %%%%%%%%%%%%%%%%%%%% First reorganization %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Reorganise particles so that the particles with same values gather
    [cand_particles_reorg, log_cand_weights_reorg, temp, new_part_ind] ...
        = reorganize_particles(cand_particles, log_cand_weights);
    cand_weights_reorg = exp(log_cand_weights_reorg);
    N_cand_reorg(t) = length(log_cand_weights_reorg);

    %%%%%%%%%%%%%%%%%%%%%%%% Forward Smoothing %%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Initialize the additive elements of the sufficient statistics
    % The quantities for parameter of the geometric distribution:
    S_1 = cell(1, Q); S_2 = cell(1, Q); S_3 = cell(Q, Q);
    S_5 = cell(1, Q); S_6 = cell(1, Q); S_7 = cell(Q, Q);

    % The quantities for alpha and beta:
    S_4 = zeros(N_reorg(t-1), N_cand_reorg(t));
    for q1 = 1:Q
        % The quantities for parameter of the geometric distribution:
        S_1{q1} = zeros(N_reorg(t-1),N_cand_reorg(t));
        S_2{q1} = zeros(N_reorg(t-1),N_cand_reorg(t));
        % The quantites for delta:
        S_5{q1} = zeros(N_reorg(t-1),N_cand_reorg(t));
        % The quantites for eta:
        S_6{q1} = zeros(N_reorg(t-1),N_cand_reorg(t));
        S_7{q1} = zeros(N_reorg(t-1),N_cand_reorg(t));
        % The quantities for the transition probability matrix for the models:
        for q2 = 1:Q
            S_3{q1, q2} = zeros(N_reorg(t-1),N_cand_reorg(t));
        end
    end

    % Initialize the transition density matrix
    F = -inf*ones(N_cand_reorg(t),N_reorg(t-1));

    for i = 1:N_cand_reorg(t)

        cd = cand_particles_reorg(i).D;
        cm = cand_particles_reorg(i).M;

        prior_params.alpha = alpha;
        prior_params.beta = beta;
        prior_params.eta = eta(cm);
        prior_params.delta = delta(cm);

        % Posterior dist. conditioned on the current particle at time t
        [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t), z(cd+1:t));
        alpha_post_curr = post_params_curr.alpha_post;
        beta_post_curr = post_params_curr.beta_post;
        eta_post_curr = post_params_curr.eta_post;
        delta_post_curr = post_params_curr.delta_post;

        % Posterior dist. conditioned on the current particle at time t-1
        [post_params_prev] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t-1), z(cd+1:t-1));
        alpha_post_prev = post_params_prev.alpha_post;
        beta_post_prev = post_params_prev.beta_post;
        eta_post_prev = post_params_prev.eta_post;
        delta_post_prev = post_params_prev.delta_post;

        for j = 1:N_reorg(t-1)

            pd = particles_reorg_prev(j).D;
            pm = particles_reorg_prev(j).M;

            if cd == t-1
                F(i, j) = log(success_prob(pm)) + log(mdl_trans_mtx(pm, cm));

                S_1{cm}(j,i) = 1; S_3{pm, cm}(j,i) = 1;
                % Alpha and beta
                S_4(j,i) = psi(alpha_post_curr) - log(beta_post_curr);
                % delta and eta
                S_5{cm}(j,i) = 1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr;
                S_6{cm}(j,i) = eta_post_curr*(alpha_post_curr/beta_post_curr);
                S_7{cm}(j,i) = (alpha_post_curr/beta_post_curr);

            elseif (cd == pd)&&(cm == pm)
                F(i, j) = log(1 - success_prob(pm));

                S_2{cm}(j,i) = 1;
                % Alpha and beta
                S_4(j,i) = psi(alpha_post_curr) - log(beta_post_curr)...
                    - (psi(alpha_post_prev) - log(beta_post_prev));
                % delta and eta
                S_5{cm}(j,i) = 1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr...
                    - (1/delta_post_prev + eta_post_prev^2*alpha_post_prev/beta_post_prev);
                S_6{cm}(j,i) = (eta_post_curr*(alpha_post_curr/beta_post_curr)...
                    - eta_post_prev*(alpha_post_prev/beta_post_prev));
                S_7{cm}(j,i) = (alpha_post_curr/beta_post_curr - alpha_post_prev/beta_post_prev);
            end
        end

    end

    %% Update the intermediate quantities:
    % Store the previous values (for the ones needed)
    T_1_prev = T_1; T_2_prev = T_2; T_3_prev = T_3;
    T_5_prev = T_5; T_6_prev = T_6; T_7_prev = T_7;
    % Initialize the new variables
    T_1 = zeros(N_cand_reorg(t),Q); T_2 = zeros(N_cand_reorg(t),Q);
    T_3 = zeros(N_cand_reorg(t),Q, Q); T_5 = zeros(N_cand_reorg(t),Q);
    T_6 = zeros(N_cand_reorg(t),Q); T_7 = zeros(N_cand_reorg(t),Q);
    for q = 1:Q
        [T_1(:,q)] = T_update(T_1_prev(:,q), S_1{q}, F, log_filt_weights_reorg_prev, w1, w2);
        [T_2(:,q)] = T_update(T_2_prev(:,q), S_2{q}, F, log_filt_weights_reorg_prev, w1, w2);
        [T_5(:,q)] = T_update(T_5_prev(:,q), S_5{q}, F, log_filt_weights_reorg_prev, w1, w2);
        [T_6(:,q)] = T_update(T_6_prev(:,q), S_6{q}, F, log_filt_weights_reorg_prev, w1, w2);
        [T_7(:,q)] = T_update(T_7_prev(:,q), S_7{q}, F, log_filt_weights_reorg_prev, w1, w2);
    end
    for q1 = 1:Q
        for q2 = 1:Q
            [T_3(:,q1,q2)] = T_update(T_3_prev(:,q1,q2), S_3{q1,q2}, F, log_filt_weights_reorg_prev, w1, w2);
        end
    end
    [T_4] = T_update(T_4, S_4, F, log_filt_weights_reorg_prev, w1, w2);

    %% Update the sufficient statistics
    for q1 = 1:Q
        for q2 = 1:Q
            SS_FS(t, 2*Q + (q1-1)*Q + q2) = cand_weights_reorg*T_3(:,q1,q2);
        end
    end
    SS_FS(t, [1:(2*Q) (Q^2 + 2*Q + 1):(Q^2 + 5*Q + 1)]) = ...
        cand_weights_reorg*[T_1 T_2 T_4 T_5 T_6 T_7];

    %%%%%%%%%%%%%%%%%%%%%%%% Maximisation step %%%%%%%%%%%%%%%%%%%%%%%%
    if t > burn_in_per
        SS_1 = SS_FS(t, 1:Q);
        SS_2 = SS_FS(t, Q+1:2*Q);
        SS_3 = reshape(SS_FS(t, (2*Q + 1): (Q^2 + 2*Q)), Q, Q)';
        SS_4 = SS_FS(t, (Q^2 + 2*Q + 1));
        SS_5 = SS_FS(t, (Q^2 + 2*Q + 2):(Q^2 + 3*Q + 1));
        SS_6 = SS_FS(t, (Q^2 + 3*Q + 2):(Q^2 + 4*Q + 1));
        SS_7 = SS_FS(t, (Q^2 + 4*Q + 2):(Q^2 + 5*Q + 1));

        % Perform parameter estimation:
        if inputs.fix_params(1)==0
            success_prob = (SS_1./(SS_1 + SS_2))';
        end
        if inputs.fix_params(2)==0
            mdl_trans_mtx = SS_3./repmat(sum(SS_3,2),1,Q);
        end
        if inputs.fix_params(3)==0
            alpha = max_over_shape_param(sum(SS_1)*log(beta) + SS_4, sum(SS_1));
        end
        if inputs.fix_params(4)==0
            beta = sum(SS_1)*alpha/sum(SS_7);
        end
        if inputs.fix_params(5)==0
            eta = (SS_6./SS_7)';
        end
        if inputs.fix_params(6)==0
            delta = (SS_1./(SS_5 - 2*eta'.*SS_6 + eta'.^2.*SS_7))';
        end

        for q = 1:Q

            seg_length_pmf_params.success_prob = success_prob(q);

            % Inner parameters for the GC_content system:
            inner_params_est.alpha = alpha;
            inner_params_est.beta = beta;
            inner_params_est.delta = delta(q);
            inner_params_est.eta = eta(q);

            model_params{q}.inner_params = inner_params_est;
            model_params{q}.seg_length_pmf_params = seg_length_pmf_params;

        end
    end
    % Store the parameter estimates over time:
    Alpha_est(t) = alpha; Beta_est(t) = beta;
    Delta_est(:,t) = delta; Eta_est(:,t) = eta;
    Success_prob_est(:,t) = success_prob;
    Mdl_trans_mtx_est(:,t) = mdl_trans_mtx(:);

    if tilda < 1
        if mod(t, 100) == 0
            figure(10);
            subplot(321);
            plot(Alpha_est(1:t));
            title('\alpha estimate');

            subplot(322);
            plot(Beta_est(1:t));
            title('\beta estimate');

            subplot(323);
            plot(Eta_est(:,1:t)');

            title('\eta estimate');
            subplot(324);
            plot(Delta_est(:,1:t)');
            title('\delta estimate');

            subplot(325);
            plot(Success_prob_est(:,1:t)');
            title('\gamma estimate');

            subplot(326);
            plot(Mdl_trans_mtx_est(:, 1:t)');
            drawnow;
        end
    end

    %%%%%%%%%%%%%%%%% Resampling to N particles %%%%%%%%%%%%%%%%%%%%%%%%%%%
    resampled_indices = multinomial_res(log_cand_weights, N);
    current_particles = cand_particles(resampled_indices);
    log_filt_weights = -log(N)*ones(N, 1);

    %%% Resample intermediate quantities along with the filter particles
    T_1 = T_1(new_part_ind(resampled_indices), :);
    T_2 = T_2(new_part_ind(resampled_indices), :);
    T_3 = T_3(new_part_ind(resampled_indices), :, :);
    T_4 = T_4(new_part_ind(resampled_indices));
    T_5 = T_5(new_part_ind(resampled_indices), :);
    T_6 = T_6(new_part_ind(resampled_indices), :);
    T_7 = T_7(new_part_ind(resampled_indices), :);

    %%% Reorganize for the forward smoothing method for the second time
    [particles_reorg, log_filt_weights_reorg, reorg_ind] ...
        = reorganize_particles(current_particles, log_filt_weights);
    T_1 = T_1(reorg_ind, :);
    T_2 = T_2(reorg_ind, :);
    T_3 = T_3(reorg_ind, :, :);
    T_4 = T_4(reorg_ind);
    T_5 = T_5(reorg_ind, :);
    T_6 = T_6(reorg_ind, :);
    T_7 = T_7(reorg_ind, :);
    N_reorg(t) = length(reorg_ind);

end

%% Outputs: 
outputs.Alpha_est = Alpha_est;
outputs.Beta_est = Beta_est;
outputs.Delta_est = Delta_est;
outputs.Eta_est = Eta_est;
outputs.Success_prob_est = Success_prob_est;
outputs.Mdl_trans_mtx_est = Mdl_trans_mtx_est;