function [outputs] = online_EM_PB_for_GC_content(y, inputs, z)

% [outputs] = online_EM_PB_for_GC_content(y, inputs, z)
%
% This function performs the online EM algorithm using the path space
% method. Bootstrap for filtering and multinomial for resampling are used.
%


% Extract the initial values:
alpha = inputs.init_params.alpha;
beta = inputs.init_params.beta;
eta = inputs.init_params.eta;
delta = inputs.init_params.delta;
success_prob = inputs.init_params.success_prob;
mdl_trans_mtx = inputs.init_params.mdl_trans_mtx;

Q = length(eta); T = length(y);

% Initialize the arrays of parameter estimates over time:
Alpha_est = zeros(1, T); Alpha_est(1) = alpha;
Beta_est = zeros(1, T); Beta_est(1) = beta;
Delta_est = zeros(Q, T); Delta_est(:,1) = delta;
Eta_est = zeros(Q, T); Eta_est(:,1) = eta;
Success_prob_est = zeros(Q, T); Success_prob_est(:,1) = success_prob;
Mdl_trans_mtx_est = zeros(Q^2, T); Mdl_trans_mtx_est(:,1) = mdl_trans_mtx(:);

% The vector of indicator for missing data problem
if nargin == 2
    z = ones(1,T);
end

model_params = cell(1,Q);
for q = 1:Q

    seg_length_pmf_name = 'geometric_pdf';
    seg_length_pmf_params.success_prob = success_prob(q);
    seg_length_pmf_params.start_pnt = 1;

    data_log_lkl_pdf_name = 'log_lkl_GC_content';

    % Inner parameters for the GC_content system:
    inner_params.alpha = alpha;
    inner_params.beta = beta;
    inner_params.delta = delta(q);
    inner_params.eta = eta(q);

    model_params{q}.seg_length_pmf_name = seg_length_pmf_name;
    model_params{q}.seg_length_pmf_params = seg_length_pmf_params;
    model_params{q}.data_log_lkl_pdf_name = data_log_lkl_pdf_name;
    model_params{q}.inner_params = inner_params;
end

% Extract the system parameters:
N = inputs.Num_of_pars;
burn_in_per = inputs.burn_in_per;
tilda = inputs.tilda;
system_type = inputs.system_type;

% Initialize particles:
current_particles(N) = struct;
for i = 1:N
    current_particles(i).D = 0; current_particles(i).M = mod(i-1, Q) + 1;
end

% Initialize the log-filtering density
log_filt_weights = -log(N)*ones(1, N);
filt_weights = exp(log_filt_weights);


% Calculate the parameters of the posterior distribution of the latent
% variable at time t=1.
alpha_post_curr = zeros(Q, 1);
beta_post_curr = zeros(Q, 1);
eta_post_curr = zeros(Q, 1);
delta_post_curr = zeros(Q, 1);
for q = 1:Q
    prior_params.alpha = alpha;
    prior_params.beta = beta;
    prior_params.eta = eta(q);
    prior_params.delta = delta(q);

    % Posterior dist. conditioned on the current particle at time t
    [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
        y(1), z(1));
    alpha_post_curr(q) = post_params_curr.alpha_post;
    beta_post_curr(q) = post_params_curr.beta_post;
    eta_post_curr(q) = post_params_curr.eta_post;
    delta_post_curr(q) = post_params_curr.delta_post;
end

%% %%%%%%%%%% Initalization for the path base method
% The quantities for parameter of the geometric distribution:
S_1_PB = zeros(N, Q); S_2_PB = zeros(N, Q);
% The quantities for the transition probability matrix for the models:
S_3_PB = ones(N, Q, Q);
% The quantities for alpha and beta:
S_4_PB = psi(alpha_post_curr(mod((0:(N-1))', Q)+1)) ...
    - log(beta_post_curr(mod((0:(N-1))', Q)+1));
% The quantites for delta:
S_5_PB = zeros(N, Q);
S_6_PB = zeros(N, Q); S_7_PB = zeros(N, Q);
for i = 1:N
    q = current_particles.M;
    S_1_PB(i, q) = 1;
    S_5_PB(i, q) = 1/delta_post_curr(q) ...
        + eta_post_curr(q)^2*alpha_post_curr(q)/beta_post_curr(q);
    S_6_PB(i, q) = eta_post_curr(q)*(alpha_post_curr(q)/beta_post_curr(q));
    S_7_PB(i, q) = (alpha_post_curr(q)/beta_post_curr(q));
end

% Initialise the sufficient statistics and store them
SS_PB = zeros(T, Q^2 + 5*Q + 1);
for q1 = 1:Q
    for q2 = 1:Q
        SS_PB(1, 2*Q + (q1-1)*Q + q2) = filt_weights*S_3_PB(:,q1,q2);
    end
end
SS_PB(1, [1:(2*Q) (Q^2 + 2*Q + 1):(Q^2 + 5*Q + 1)]) = ...
    filt_weights*[S_1_PB S_2_PB S_4_PB S_5_PB S_6_PB S_7_PB];


%%%%%%%%%%%%%%%%%%%%%%%%%% Recursion over time: %%%%%%%%%%%%%%%%%%%%%%%%%%%
for t = 2:T

    if mod(t, 100) == 0
        fprintf('%d ', t);
        if mod(t, 1000) == 0
            fprintf('\n');
        end
    end
    % Step size

    if t <= burn_in_per
        gam = 0.01;
    else
        gam = (t)^(-tilda);
    end
    w1 = gam; w2 = 1 - gam;

    prev_particles = current_particles;

    %%%%%%%%%%%%%%%%%%%%% Bootstrap Filtering %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Bootstrap method:
    cand_particles = prev_particles;
    log_cand_weights = zeros(1, N);

    for i = 1:N
        % Sample from the Proposal density
        M_prev = prev_particles(i).M;
        D_prev = prev_particles(i).D;
        if rand < success_prob(M_prev)
            cand_particles(i).D = t-1;
            cand_particles(i).M = randgen(1:Q, 1, mdl_trans_mtx(M_prev, :));
            % Incremental weight:
            inc_weight = log_lkl_y(y(t), z(t), model_params{cand_particles(i).M}, system_type);
        else
            % Incremental weight:
            inc_weight = log_lkl_y(y(D_prev +1:t), z(D_prev +1:t),model_params{M_prev}, system_type)...
                - log_lkl_y(y(D_prev+1:t-1), z(D_prev+1:t-1), model_params{M_prev}, system_type);
        end
        log_cand_weights(i) = inc_weight;
    end
    log_cand_weights = log_cand_weights - log_sum(log_cand_weights);
    cand_weights = exp(log_cand_weights);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%% Sufficient Statistics update %%%%%%%%%%%%%%%%%
            
    S_1_PB = w2*S_1_PB; S_2_PB = w2*S_2_PB; S_3_PB = w2*S_3_PB;
    S_4_PB = w2*S_4_PB; S_5_PB = w2*S_5_PB; S_6_PB = w2*S_6_PB;
    S_7_PB = w2*S_7_PB;
    for i = 1:N      
        cd = cand_particles(i).D;
        cm = cand_particles(i).M;      
        pd = prev_particles(i).D;
        pm = prev_particles(i).M;
        
        prior_params.alpha = alpha;
        prior_params.beta = beta;
        prior_params.eta = eta(cm);
        prior_params.delta = delta(cm);
        
        % Posterior dist. conditioned on the current particle at time t
        [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t), z(cd+1:t));
        alpha_post_curr = post_params_curr.alpha_post;
        beta_post_curr = post_params_curr.beta_post;
        eta_post_curr = post_params_curr.eta_post;
        delta_post_curr = post_params_curr.delta_post;
        
        % Posterior dist. conditioned on the current particle at time t-1
        [post_params_prev] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t-1), z(cd+1:t-1));
        alpha_post_prev = post_params_prev.alpha_post;
        beta_post_prev = post_params_prev.beta_post;
        eta_post_prev = post_params_prev.eta_post;
        delta_post_prev = post_params_prev.delta_post;

        if cd == t-1                        
            S_1_PB(i, cm) = S_1_PB(i, cm) + w1*1; 
            S_3_PB(i, pm, cm) = S_3_PB(i, pm, cm) + w1*1;
            % Alpha and beta
            S_4_PB(i) = S_4_PB(i) + w1*(psi(alpha_post_curr) - log(beta_post_curr));
            % delta and eta
            S_5_PB(i, cm) = S_5_PB(i, cm) + w1*(1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr);
            S_6_PB(i, cm) = S_6_PB(i, cm) + w1*eta_post_curr*(alpha_post_curr/beta_post_curr);
            S_7_PB(i, cm) = S_7_PB(i, cm) + w1*(alpha_post_curr/beta_post_curr);
        elseif (cd == pd)&&(cm == pm)
            S_2_PB(i, cm) = S_2_PB(i, cm) + w1*1;
            % Alpha and beta
            S_4_PB(i) = S_4_PB(i) + w1*(psi(alpha_post_curr) - log(beta_post_curr)...
                - (psi(alpha_post_prev) - log(beta_post_prev)));
            % delta and eta
            S_5_PB(i, cm) = S_5_PB(i, cm) + w1*(1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr...
                - (1/delta_post_prev + eta_post_prev^2*alpha_post_prev/beta_post_prev));
            S_6_PB(i, cm) = S_6_PB(i, cm) + w1*(eta_post_curr*(alpha_post_curr/beta_post_curr)...
                - eta_post_prev*(alpha_post_prev/beta_post_prev));
            S_7_PB(i, cm) = S_7_PB(i, cm) + w1*((alpha_post_curr/beta_post_curr - alpha_post_prev/beta_post_prev));
        end
    end

    %% Update the sufficient statistics
    for q1 = 1:Q
        for q2 = 1:Q
            SS_PB(t, 2*Q + (q1-1)*Q + q2) = cand_weights*S_3_PB(:,q1,q2);
        end
    end
    SS_PB(t, [1:(2*Q) (Q^2 + 2*Q + 1):(Q^2 + 5*Q + 1)]) = ...
        cand_weights*[S_1_PB S_2_PB S_4_PB S_5_PB S_6_PB S_7_PB];
    
    %%%%%%%%%%%%%%%%%%%%%%%% Maximisation step %%%%%%%%%%%%%%%%%%%%%%%%
    if t > burn_in_per
        SS_1 = SS_PB(t, 1:Q);
        SS_2 = SS_PB(t, Q+1:2*Q);
        SS_3 = reshape(SS_PB(t, (2*Q + 1): (Q^2 + 2*Q)), Q, Q)';
        SS_4 = SS_PB(t, (Q^2 + 2*Q + 1));
        SS_5 = SS_PB(t, (Q^2 + 2*Q + 2):(Q^2 + 3*Q + 1));
        SS_6 = SS_PB(t, (Q^2 + 3*Q + 2):(Q^2 + 4*Q + 1));
        SS_7 = SS_PB(t, (Q^2 + 4*Q + 2):(Q^2 + 5*Q + 1));
 
        % Perform parameter estimation:
        if inputs.fix_params(1)==0
            success_prob = (SS_1./(SS_1 + SS_2))';
        end
        if inputs.fix_params(2)==0
            mdl_trans_mtx = SS_3./repmat(sum(SS_3,2),1,Q);
        end
        if inputs.fix_params(3)==0
            alpha = max_over_shape_param(sum(SS_1)*log(beta) + SS_4, sum(SS_1));
        end
        if inputs.fix_params(4)==0
            beta = sum(SS_1)*alpha/sum(SS_7);
        end
        if inputs.fix_params(5)==0
            eta = (SS_6./SS_7)';
        end
        if inputs.fix_params(6)==0
            delta = (SS_1./(SS_5 - 2*eta'.*SS_6 + eta'.^2.*SS_7))';
        end

        for q = 1:Q

            seg_length_pmf_params.success_prob = success_prob(q);

            % Inner parameters for the GC_content system:
            inner_params_est.alpha = alpha;
            inner_params_est.beta = beta;
            inner_params_est.delta = delta(q);
            inner_params_est.eta = eta(q);

            model_params{q}.inner_params = inner_params_est;
            model_params{q}.seg_length_pmf_params = seg_length_pmf_params;

        end
    end
    % Store the parameter estimates over time:
    Alpha_est(t) = alpha; Beta_est(t) = beta;
    Delta_est(:,t) = delta; Eta_est(:,t) = eta;
    Success_prob_est(:,t) = success_prob;
    Mdl_trans_mtx_est(:,t) = mdl_trans_mtx(:);

    if tilda < 1
        if mod(t, 100) == 0
            figure(11);
            subplot(321);
            plot(Alpha_est(1:t));
            title('\alpha estimate');

            subplot(322);
            plot(Beta_est(1:t));
            title('\beta estimate');

            subplot(323);
            plot(Eta_est(:,1:t)');

            title('\eta estimate');
            subplot(324);
            plot(Delta_est(:,1:t)');
            title('\delta estimate');

            subplot(325);
            plot(Success_prob_est(:,1:t)');
            title('\gamma estimate');

            subplot(326);
            plot(Mdl_trans_mtx_est(:, 1:t)');
            drawnow;
        end
    end

    %%%%%%%%%%%%%%%%% Resampling to N particles %%%%%%%%%%%%%%%%%%%%%%%%%%%
    resampled_indices = multinomial_res(log_cand_weights, N);
    current_particles = cand_particles(resampled_indices);
    
    %%% Resample sufficient statistics along with the paths 
    S_1_PB = S_1_PB(resampled_indices, :);
    S_2_PB = S_2_PB(resampled_indices, :);
    S_3_PB = S_3_PB(resampled_indices, :, :);
    S_4_PB = S_4_PB(resampled_indices);
    S_5_PB = S_5_PB(resampled_indices, :);
    S_6_PB = S_6_PB(resampled_indices, :);
    S_7_PB = S_7_PB(resampled_indices, :);

end

%% Outputs: 
outputs.Alpha_est = Alpha_est;
outputs.Beta_est = Beta_est;
outputs.Delta_est = Delta_est;
outputs.Eta_est = Eta_est;
outputs.Success_prob_est = Success_prob_est;
outputs.Mdl_trans_mtx_est = Mdl_trans_mtx_est;