function [outputs] = online_EM_for_GC_content(y, inputs, z)

% [outputs] = online_EM_for_GC_content(y, inputs, z)
%
% This function performs the online EM algorithm for changepoint models
% by using the forward smoothing method.
%


% Extract the initial values:
alpha_est = inputs.init_params.alpha;
beta_est = inputs.init_params.beta;
eta_est = inputs.init_params.eta;
delta_est = inputs.init_params.delta;
success_prob_est = inputs.init_params.success_prob;
mdl_trans_mtx_est = inputs.init_params.mdl_trans_mtx;

Q = length(eta_est); T = length(y);
% The vector of indicator for missing data problem
if nargin == 2
    z = ones(1,T);
end
% Initialize the arrays of parameter estimates over time:
Alpha_est = zeros(1, T); Alpha_est(1) = alpha_est;
Beta_est = zeros(1, T); Beta_est(1) = beta_est;
Delta_est = zeros(Q, T); Delta_est(:,1) = delta_est;
Eta_est = zeros(Q, T); Eta_est(:,1) = eta_est;
Success_prob_est = zeros(Q, T); Success_prob_est(:,1) = success_prob_est;
Mdl_trans_mtx_est = zeros(Q^2, T); Mdl_trans_mtx_est(:,1) = mdl_trans_mtx_est(:);

model_params_est = cell(1,Q);
for q = 1:Q

    seg_length_pmf_name = 'geometric_pdf';
    seg_length_pmf_params.success_prob = success_prob_est(q);
    seg_length_pmf_params.start_pnt = 1;

    data_log_lkl_pdf_name = 'log_lkl_GC_content';

    % Inner parameters for the GC_content system:
    inner_params_est.alpha = alpha_est;
    inner_params_est.beta = beta_est;
    inner_params_est.delta = delta_est(q);
    inner_params_est.eta = eta_est(q);

    model_params_est{q}.seg_length_pmf_name = seg_length_pmf_name;
    model_params_est{q}.seg_length_pmf_params = seg_length_pmf_params;
    model_params_est{q}.data_log_lkl_pdf_name = data_log_lkl_pdf_name;
    model_params_est{q}.inner_params = inner_params_est;

end

% Extract the system parameters:
tilda = inputs.tilda;
burn_in_per = inputs.burn_in_per;
weight_thr = inputs.weight_thr;
num_thr = inputs.num_thr;
system_type = inputs.system_type;

% Initialize particles:
current_particles(Q) = struct;
for q = 1:Q
    current_particles(q).D = 0; current_particles(q).M = q;
end

% Initialize the vector for the number of particles:
N = zeros(1, T); N(1) = Q;

% Initialize the log-filtering density
log_filt_weights = -log(Q)*ones(1, Q);

% Assign current estimates as the previous estimates for the next time.
alpha_est_prev = alpha_est;
beta_est_prev = beta_est;
delta_est_prev = delta_est;
eta_est_prev = eta_est;

% Calculate the parameters of the posterior distribution of the latent
% variable at time t=1.
alpha_post_curr = zeros(Q, 1);
beta_post_curr = zeros(Q, 1);
eta_post_curr = zeros(Q, 1);
delta_post_curr = zeros(Q, 1);
for q = 1:Q
    prior_params.alpha = alpha_est_prev;
    prior_params.beta = beta_est_prev;
    prior_params.eta = eta_est_prev(q);
    prior_params.delta = delta_est_prev(q);

    % Posterior dist. conditioned on the current particle at time t
    [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
        y(1), z(1));        
    alpha_post_curr(q) = post_params_curr.alpha_post;
    beta_post_curr(q) = post_params_curr.beta_post;
    eta_post_curr(q) = post_params_curr.eta_post;
    delta_post_curr(q) = post_params_curr.delta_post;
end
% Initalize the intermediate quantities needed for on-line EM

% The quantities for parameter of the geometric distribution:
T_1 = ones(N(1),Q); T_2 = zeros(N(1),Q);
% The quantities for the transition probability matrix for the models:
T_3 = ones(N(1),Q, Q);
% The quantities for alpha:
T_4 = ones(N(1),1); T_5 = ones(N(1),1);
T_6 = psi(alpha_post_curr) - log(beta_post_curr);
% The quantities for beta:
T_7 = ones(N(1),1); T_8 = alpha_post_curr./beta_post_curr;
% The quantites for delta:
T_9 = ones(N(1),Q); T_10 = zeros(N(1),Q); 
T_11 = zeros(N(1),Q); T_12 = zeros(N(1),Q);
for q = 1:Q
    T_10(q, q) = 1/delta_post_curr(q)+eta_post_curr(q)^2*alpha_post_curr(q)/beta_post_curr(q);
    T_11(q, q) = eta_post_curr(q)*(alpha_post_curr(q)/beta_post_curr(q));
    T_12(q, q) = (alpha_post_curr(q)/beta_post_curr(q));
end

%% Recursion over time:
for t = 2:T
    
    % disp([t z(t)]);
    % Display t
    if mod(t, 100)==0
        disp(t);
    end
    % Step size
    if tilda == 1
        w1 = 1; w2 = 1;
    else
        if t <= burn_in_per
            gam = 0.01;
        else
            gam = (t)^(-tilda);
        end
        w1 = gam; w2 = 1 - gam;
    end

    % Keep the previous filtering density
    log_filt_weights_prev = log_filt_weights;
    prev_particles = current_particles;

    %%
    % Initialize the candidate weights and particles:
    cand_weights = zeros(1, N(t-1)+Q);
    % Keep the previous parameters as the current candidates
    cand_particles(1:N(t-1)) = prev_particles(1:N(t-1));
    for q = 1:Q
        cand_particles(N(t-1)+q).D = t-1;
        cand_particles(N(t-1)+q).M = q;
    end

    % Filtering density update
    for i = 1:N(t-1)+Q
        % Update their candidate filtering densities
        cand_weights(i) = filt_dens_update_d_and_m(cand_particles(i),...
            prev_particles, y(1:t), z(1:t), log_filt_weights_prev, ...
            model_params_est, mdl_trans_mtx_est, t, system_type);
    end

    %% Resample the particles
    
    resampled_indices = resample_weights(cand_weights, weight_thr, num_thr);
    N(t) = length(resampled_indices);
    current_particles = cand_particles(resampled_indices);
    % Construct the approximate filtering density according to resampling
    log_filt_weights = cand_weights(resampled_indices);
    log_filt_weights = log_filt_weights - log_sum(log_filt_weights);

%     [resampled_indices, res_weights] = stratified_res(cand_weights, weight_thr, 1);
%     N(t) = length(resampled_indices);
%     current_particles = cand_particles(resampled_indices);
%     log_filt_weights = res_weights - log_sum(res_weights);

    %% Parameter Estimation
    % Initialize the additive elements of the sufficient statistics
    % The quantities for parameter of the geometric distribution:
    S_1 = cell(1, Q);
    for q = 1:Q
        S_1{q} = zeros(N(t-1),N(t));
    end
    S_2 = cell(1, Q);
    for q = 1:Q
        S_2{q} = zeros(N(t-1),N(t));
    end
    % The quantities for the transition probability matrix for the models:
    S_3 = cell(Q, Q);
    for q1 = 1:Q
        for q2 = 1:Q
            S_3{q1, q2} = zeros(N(t-1),N(t));
        end
    end
    % The quantities for alpha:
    S_4 = zeros(N(t-1), N(t)); S_5 = zeros(N(t-1), N(t));
    S_6 = zeros(N(t-1), N(t));
    % The quantities for beta:
    S_7 = zeros(N(t-1), N(t)); S_8 = zeros(N(t-1), N(t));
    % The quantites for delta:
    S_9 = cell(1, Q);
    for q = 1:Q
        S_9{q} = zeros(N(t-1),N(t));
    end
    S_10 = cell(1, Q);
    for q = 1:Q
        S_10{q} = zeros(N(t-1),N(t));
    end
    % The quantites for eta:
    S_11 = cell(1, Q);
    for q = 1:Q
        S_11{q} = zeros(N(t-1),N(t));
    end
    S_12 = cell(1, Q);
    for q = 1:Q
        S_12{q} = zeros(N(t-1),N(t));
    end

    % Initialize the transition density matrix
    F = -inf*ones(N(t),N(t-1));

    for i = 1:N(t)
        
        cd = current_particles(i).D;
        cm = current_particles(i).M;

        prior_params.alpha = alpha_est_prev;
        prior_params.beta = beta_est_prev;
        prior_params.eta = eta_est_prev(cm);
        prior_params.delta = delta_est_prev(cm);
        
        % Posterior dist. conditioned on the current particle at time t
        [post_params_curr] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t), z(cd+1:t));        
        alpha_post_curr = post_params_curr.alpha_post;
        beta_post_curr = post_params_curr.beta_post;
        eta_post_curr = post_params_curr.eta_post;
        delta_post_curr = post_params_curr.delta_post;
        
        % Posterior dist. conditioned on the current particle at time t-1
        [post_params_prev] = post_dens_in_segment(prior_params, system_type,...
            y(cd+1:t-1), z(cd+1:t-1));
        alpha_post_prev = post_params_prev.alpha_post;
        beta_post_prev = post_params_prev.beta_post;
        eta_post_prev = post_params_prev.eta_post;
        delta_post_prev = post_params_prev.delta_post;
        
        for j = 1:N(t-1)

            % Calculate the log-transition density
            F(i,j) = trans_dens_d_and_m(current_particles(i), ...
                prev_particles(j), model_params_est, mdl_trans_mtx_est, t);

            pd = prev_particles(j).D;
            pm = prev_particles(j).M;
            if cd == t-1
                S_1{pm}(j,i) = 1; S_3{pm, cm}(j,i) = 1;
                % Alpha
                S_4(j,i) = 1; S_5(j,i) = 1;
                S_6(j,i) = psi(alpha_post_curr) - log(beta_post_curr);
                % Beta
                S_7(j,i) = 1; 
                S_8(j,i) = alpha_post_curr/beta_post_curr;
                % delta and eta
                S_9{cm}(j,i) = 1;
                S_10{cm}(j,i) = 1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr;
                S_11{cm}(j,i) = eta_post_curr*(alpha_post_curr/beta_post_curr);
                S_12{cm}(j,i) = (alpha_post_curr/beta_post_curr);

            elseif (cd == pd)&&(cm == pm)

                S_2{cm}(j,i) = 1;
                % Alpha
                S_6(j,i) = psi(alpha_post_curr) - log(beta_post_curr)...
                    - (psi(alpha_post_prev) - log(beta_post_prev));
                % Beta
                S_8(j,i) = alpha_post_curr/beta_post_curr...
                    - alpha_post_prev/beta_post_prev;
                % delta and eta
                S_10{cm}(j,i) = 1/delta_post_curr + eta_post_curr^2*alpha_post_curr/beta_post_curr...
                    - (1/delta_post_prev + eta_post_prev^2*alpha_post_prev/beta_post_prev);
                S_11{cm}(j,i) = (eta_post_curr*(alpha_post_curr/beta_post_curr)...
                    - eta_post_prev*(alpha_post_prev/beta_post_prev));
                S_12{cm}(j,i) = (alpha_post_curr/beta_post_curr - alpha_post_prev/beta_post_prev);
                
            end
        end

    end

    %% Update the intermediate quantities:
    % Store the previous values (for the ones needed)
    T_1_prev = T_1; T_2_prev = T_2; T_3_prev = T_3; T_9_prev = T_9;
    T_10_prev = T_10; T_11_prev = T_11; T_12_prev = T_12;
    % Initialize the new variables
    T_1 = zeros(N(t),Q); T_2 = zeros(N(t),Q); T_3 = zeros(N(t),Q, Q);
    T_9 = zeros(N(t),Q); T_10 = zeros(N(t),Q); T_11 = zeros(N(t),Q);
    T_12 = zeros(N(t),Q);
    for q = 1:Q
        [T_1(:,q)] = T_update(T_1_prev(:,q), S_1{q}, F, log_filt_weights_prev, w1, w2);
        [T_2(:,q)] = T_update(T_2_prev(:,q), S_2{q}, F, log_filt_weights_prev, w1, w2);
        [T_9(:,q)] = T_update(T_9_prev(:,q), S_9{q}, F, log_filt_weights_prev, w1, w2);
        [T_10(:,q)] = T_update(T_10_prev(:,q), S_10{q}, F, log_filt_weights_prev, w1, w2);
        [T_11(:,q)] = T_update(T_11_prev(:,q), S_11{q}, F, log_filt_weights_prev, w1, w2);
        [T_12(:,q)] = T_update(T_12_prev(:,q), S_12{q}, F, log_filt_weights_prev, w1, w2);
    end
    for q1 = 1:Q
        for q2 = 1:Q
            [T_3(:,q1,q2)] = T_update(T_3_prev(:,q1,q2), S_3{q1,q2}, F, log_filt_weights_prev, w1, w2);
        end
    end
    [T_4] = T_update(T_4, S_4, F, log_filt_weights_prev, w1, w2);
    [T_5] = T_update(T_5, S_5, F, log_filt_weights_prev, w1, w2);
    [T_6] = T_update(T_6, S_6, F, log_filt_weights_prev, w1, w2);
    [T_7] = T_update(T_7, S_7, F, log_filt_weights_prev, w1, w2);
    [T_8] = T_update(T_8, S_8, F, log_filt_weights_prev, w1, w2);

    
    % Update the sufficient statistics
    if t > burn_in_per
        
        filt_weights = exp(log_filt_weights);
        SS_1 = filt_weights*T_1; SS_2 = filt_weights*T_2;
        SS_3 = zeros(Q,Q);
        for q1 = 1:Q
            for q2 = 1:Q
                SS_3(q1,q2) = filt_weights*T_3(:,q1,q2);
            end
        end
        SS_4 = filt_weights*T_4; SS_5 = filt_weights*T_5;
        SS_6 = filt_weights*T_6; SS_7 = filt_weights*T_7;
        SS_8 = filt_weights*T_8; SS_9 = filt_weights*T_9;
        SS_10 = filt_weights*T_10; SS_11 = filt_weights*T_11;
        SS_12 = filt_weights*T_12;     

        % Perform parameter estimation:
        if inputs.fix_params(1)==0
            success_prob_est = (SS_1./(SS_1 + SS_2))';
        end
        if inputs.fix_params(2)==0
            mdl_trans_mtx_est = SS_3./repmat(sum(SS_3,2),1,Q);
        end
        if inputs.fix_params(3)==0
            alpha_est = max_over_shape_param(SS_4*log(beta_est) + SS_6, SS_5);
        end
        if inputs.fix_params(4)==0
            beta_est = SS_7*alpha_est_prev/SS_8;
        end
        if inputs.fix_params(5)==0
            eta_est = (SS_11./SS_12)';
        end
        if inputs.fix_params(6)==0
            delta_est = (SS_9./(SS_10 - 2*eta_est'.*SS_11 + eta_est'.^2.*SS_12))';
        end
        
        for q = 1:Q
            
            seg_length_pmf_params.success_prob = success_prob_est(q);
                       
            % Inner parameters for the GC_content system:
            inner_params_est.alpha = alpha_est;
            inner_params_est.beta = beta_est;
            inner_params_est.delta = delta_est(q);
            inner_params_est.eta = eta_est(q);

            model_params_est{q}.inner_params = inner_params_est;
            model_params_est{q}.seg_length_pmf_params = seg_length_pmf_params;

        end
        % Assign current estimates as the previous estimates for the next time.
        alpha_est_prev = alpha_est;
        beta_est_prev = beta_est;
        delta_est_prev = delta_est;
        eta_est_prev = eta_est;
        
    end

    % Store the parameter estimates over time:
    Alpha_est(t) = alpha_est; Beta_est(t) = beta_est;
    Delta_est(:,t) = delta_est; Eta_est(:,t) = eta_est;
    Success_prob_est(:,t) = success_prob_est;
    Mdl_trans_mtx_est(:,t) = mdl_trans_mtx_est(:);

    if tilda < 1
        if mod(t, 100) == 0
            figure(10);
            subplot(321)
            plot(Alpha_est(1:t));
            title('\alpha estimate');

            subplot(322)
            plot(Beta_est(1:t));
            title('\beta estimate');
            
            subplot(323)
            plot(Eta_est(:,1:t)');

            title('\eta estimate');
            subplot(324)
            plot(Delta_est(:,1:t)');
            title('\delta estimate');

            subplot(325)
            plot(Success_prob_est(:,1:t)');
            title('\gamma estimate');

            subplot(326)
            plot(Mdl_trans_mtx_est(:, 1:t)');
            drawnow;
        end
    end
end

%% Outputs
outputs.Alpha_est = Alpha_est;
outputs.Beta_est = Beta_est;
outputs.Delta_est = Delta_est;
outputs.Eta_est = Eta_est;
outputs.Success_prob_est = Success_prob_est;
outputs.Mdl_trans_mtx_est = Mdl_trans_mtx_est;