function [simulated_system] = simul_chp_system(model_params, mdl_init_dist, mdl_trans_mtx, T, system_type)

% [simulated_system] = simul_chp_system(model_params, mdl_init_dist, mdl_trans_mtx, T, system_type)
%
% This function simulates all of the variables of a changepoint system, 
% specified by the input 'system_type'. The needed model parameters are 
% given in 'model_params'. Initial distribution of models 'mdl_int_dist', 
% the transition probability matrix for the models 'mdl_trans_mtx', and the
% length of the process are other needed inputs. 
%
% The output 'system_data' contains the simulation values of the following 
% random variables:
%
% C(i): The time of the i^th changepoint,
% D(t): The most recent changepoint time prior to time t,
% Z(i): Moden number at the i^th segment,
% M(t): Model number at time t. 
% x: Latent variables 
% y: Observable data
%


%% Simulate the changepoints and the model numbers:
[C, D, M, Z] = simul_chps_and_models(model_params, mdl_init_dist, mdl_trans_mtx, T);

%% Simulate the latent and the observable data:
Num_of_segments = length(C);
x = cell(1, Num_of_segments);
y = zeros(1, T);

for i = 1:Num_of_segments

    current_model = Z(i);
    if i < Num_of_segments
        segment_length = C(i+1)-C(i);
    else
        segment_length = T-C(i);
    end
    outputs = simul_data_for_one_segment(model_params{current_model}, segment_length, system_type);
    y(C(i)+1:C(i)+segment_length) = outputs.obs_data;
    x{i} = outputs.latent_data;
    
end

%% Store the outputs:
simulated_system.C = C;
simulated_system.D = D;
simulated_system.Z = Z;
simulated_system.M = M;
simulated_system.x = x;
simulated_system.y = y;