function [C, D, M, Z] = simul_chps_and_models(model_params, mdl_init_dist, mdl_trans_mtx, T)

% [C, D, M, Z] = simul_chps_and_models(model_params, mdl_init_dist,
% mdl_trans_mtx, T)
%
% This function simulates the common variables of any system with
% changepoints, independent of the inner system in the segments. The needed
% model parameters are given in 'model_params'. Initial distribution of
% models 'mdl_int_dist', the transition probability matrix for the
% models 'mdl_trans_mtx', and the length of the process are other needed
% inputs. The output is the simulation of the following random variables:
%
% C(i): The time of the i^th changepoint,
% D(t): The most recent changepoint time prior to time t,
% Z(i): Moden number at the i^th segment,
% M(t): Model number at time t. 
%


% Number of possible models:
Q = length(mdl_init_dist);

% First changepoint time is 0
t = 1;
num_of_chp = 1;
% The most recent changepoint time prior to time t:
D(t) = 0;
% The time of the num_of_chp^th changepoint
C(num_of_chp) = 0;
% The model at time 1 is randomly selected using the initial distribution:
M(t) = randgen((1:Q), 1, mdl_init_dist);
% Model number in the num_of_chp^th segment
Z(num_of_chp) = M(t);

while t < T
    % Calculate the probability of the event that next changepoint occurs
    % at time t+1 given that the most recent changepoint time is D(t):
    
    % First, extract the pmf for the length of the current segment:
    pmf = str2func(model_params{M(t)}.seg_length_pmf_name);
    % Obtain the parameters for that pmf:
    params = model_params{M(t)}.seg_length_pmf_params;
    
    % Calculate the needed values of the pmf and the corresponding cmf:
    G = sum(pmf(1:t-D(t)-1, params));
    g = pmf(t-D(t), params);
    % Now; calculate the mentioned probability:
    p = g/(1-G);
    if rand < p
        % Changepoint occurs at time t
        num_of_chp = num_of_chp + 1;
        % The most recent changepoint prior to time t+1 is t:
        D(t+1) = t;
        % The time of the num_of_chp^th changepoint is t
        C(num_of_chp) = t;
        
        % New model for time t+1
        M(t+1) = randgen((1:Q), 1, mdl_trans_mtx(M(t),:));
        Z(num_of_chp) = M(t+1); 
    else
        % No changepoint at time t
        D(t+1) = D(t);
        M(t+1) = M(t);
    end
    % Increase time by 1
    t = t+1;
end